@@ -297,8 +297,6 @@ static int wcmInitAxes(DeviceIntPtr pWcm)
 		label = None;
 		mode = Absolute;
 		min_res = max_res = res = 1;
-		min = 0;
-		max = 1;
 
 		min = MIN_PAD_RING;
 		max = MAX_PAD_RING;
@@ -480,7 +478,7 @@ Bool wcmIsWacomDevice (char* fname)
 char *wcmEventAutoDevProbe (InputInfoPtr pInfo)
 {
 	/* We are trying to find the right eventX device */
-	int i, wait = 0;
+	int i = 0, wait = 0;
 	const int max_wait = 2000;
 
 	/* If device is not available after Resume, wait some ms */
@@ -755,9 +753,9 @@ static void wcmDevClose(InputInfoPtr pInfo)
 
 	if (pInfo->fd >= 0)
 	{
-		pInfo->fd = -1;
 		if (!--common->fd_refs)
 			wcmClose(pInfo);
+		pInfo->fd = -1;
 	}
 }
 
