@@ -86,6 +86,7 @@ enum prop_flags {
 typedef struct _param
 {
 	const char *name;	/* param name as specified by the user */
+	const char *x11name;    /* param name used in xorg.conf */
 	const char *desc;	/* description */
 	const char *prop_name;	/* property name */
 	const int prop_format;	/* property format */
@@ -120,6 +121,7 @@ static param_t parameters[] =
 {
 	{
 		.name = "Area",
+		.x11name = "Area",
 		.desc = "Valid tablet area in device coordinates. ",
 		.prop_name = WACOM_PROP_TABLET_AREA,
 		.prop_format = 32,
@@ -136,6 +138,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "ToolDebugLevel",
+		.x11name = "DebugLevel",
 		.desc = "Level of debugging trace for individual tools "
 		"(default is 0 [off]). ",
 		.prop_name = WACOM_PROP_DEBUGLEVELS,
@@ -145,6 +148,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "TabletDebugLevel",
+		.x11name = "CommonDBG",
 		.desc = "Level of debugging statements applied to shared "
 		"code paths between all tools "
 		"associated with the same tablet (default is 0 [off]). ",
@@ -155,6 +159,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "Suppress",
+		.x11name = "Suppress",
 		.desc = "Number of points trimmed (default is 2). ",
 		.prop_name = WACOM_PROP_SAMPLE,
 		.prop_format = 32,
@@ -163,6 +168,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "RawSample",
+		.x11name = "RawSample",
 		.desc = "Number of raw data used to filter the points "
 		"(default is 4). ",
 		.prop_name = WACOM_PROP_SAMPLE,
@@ -172,6 +178,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "PressureCurve",
+		.x11name = "PressCurve",
 		.desc = "Bezier curve for pressure (default is 0 0 100 100 [linear]). ",
 		.prop_name = WACOM_PROP_PRESSURECURVE,
 		.prop_format = 32,
@@ -180,6 +187,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "Mode",
+		.x11name = "Mode",
 		.desc = "Switches cursor movement mode (default is absolute). ",
 		.arg_count = 1,
 		.set_func = set_mode,
@@ -187,6 +195,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "TabletPCButton",
+		.x11name = "TPCButton",
 		.desc = "Turns on/off Tablet PC buttons "
 		"(default is off for regular tablets, "
 		"on for Tablet PC). ",
@@ -198,6 +207,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "Touch",
+		.x11name = "Touch",
 		.desc = "Turns on/off Touch events (default is on). ",
 		.prop_name = WACOM_PROP_TOUCH,
 		.prop_format = 8,
@@ -216,6 +226,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "Gesture",
+		.x11name = "Gesture",
 		.desc = "Turns on/off multi-touch gesture events "
 		"(default is on). ",
 		.prop_name = WACOM_PROP_ENABLE_GESTURE,
@@ -226,6 +237,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "ZoomDistance",
+		.x11name = "ZoomDistance",
 		.desc = "Minimum distance for a zoom gesture "
 		"(default is 50). ",
 		.prop_name = WACOM_PROP_GESTURE_PARAMETERS,
@@ -235,6 +247,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "ScrollDistance",
+		.x11name = "ScrollDistance",
 		.desc = "Minimum motion before sending a scroll gesture "
 		"(default is 20). ",
 		.prop_name = WACOM_PROP_GESTURE_PARAMETERS,
@@ -244,6 +257,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "TapTime",
+		.x11name = "TapTime",
 		.desc = "Minimum time between taps for a right click "
 		"(default is 250). ",
 		.prop_name = WACOM_PROP_GESTURE_PARAMETERS,
@@ -253,6 +267,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "CursorProximity",
+		.x11name = "CursorProx",
 		.desc = "Sets cursor distance for proximity-out "
 		"in distance from the tablet "
 		"(default is 10 for Intuos series, "
@@ -264,6 +279,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "Rotate",
+		.x11name = "Rotate",
 		.desc = "Sets the rotation of the tablet. "
 		"Values = none, cw, ccw, half (default is none). ",
 		.prop_name = WACOM_PROP_ROTATION,
@@ -373,6 +389,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "Threshold",
+		.x11name = "Threshold",
 		.desc = "Sets tip/eraser pressure threshold "
 		"(default is 27). ",
 		.prop_name = WACOM_PROP_PRESSURE_THRESHOLD,
@@ -428,6 +445,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "BindToSerial",
+		.x11name = "Serial",
 		.desc = "Binds this device to the serial number.",
 		.prop_name = WACOM_PROP_SERIAL_BIND,
 		.prop_format = 32,
@@ -445,6 +463,7 @@ static param_t parameters[] =
 	},
 	{
 		.name = "PressureRecalibration",
+		.x11name = "PressureRecalibration",
 		.desc = "Turns on/off Tablet pressure recalibration",
 		.prop_name = WACOM_PROP_PRESSURE_RECAL,
 		.prop_format = 8,
@@ -519,11 +538,11 @@ is_deprecated_parameter(const char *name)
 
 	if (is_deprecated)
 	{
-		printf("Parameter '%s' is no longer in use. ", name);
+		fprintf(stderr, "Parameter '%s' is no longer in use. ", name);
 		if (d->replacement != NULL)
-			printf("It was replaced with '%s'.\n", d->replacement);
+			fprintf(stderr, "It was replaced with '%s'.\n", d->replacement);
 		else
-			printf("Its use has been deprecated.\n");
+			fprintf(stderr, "Its use has been deprecated.\n");
 	}
 
 	return is_deprecated;
@@ -531,8 +550,8 @@ is_deprecated_parameter(const char *name)
 }
 
 struct modifier {
-	char *name;
-	char *converted;
+	const char *name;
+	const char *converted;
 };
 
 static struct modifier modifiers[] = {
@@ -642,6 +661,7 @@ static param_t* find_parameter(char *name)
 	return param->name ? param : NULL;
 }
 
+static void print_value(param_t *param, const char *msg, ...) __attribute__((format(printf, 2, 3)));
 static void print_value(param_t *param, const char *msg, ...)
 {
 	va_list va_args;
@@ -649,11 +669,17 @@ static void print_value(param_t *param, const char *msg, ...)
 	switch(param->printformat)
 	{
 		case FORMAT_XORG_CONF:
-			printf("Option \"%s\" \"", param->name);
+			if (param->prop_flags & PROP_FLAG_READONLY)
+				break;
+			if (!param->x11name)
+				break;
+			printf("Option \"%s\" \"", param->x11name);
 			vprintf(msg, va_args);
 			printf("\"\n");
 			break;
 		case FORMAT_SHELL:
+			if (param->prop_flags & PROP_FLAG_READONLY)
+				break;
 			printf("xsetwacom set \"%s\" \"%s\" \"",
 					param->device_name, param->name);
 			vprintf(msg, va_args);
@@ -669,6 +695,36 @@ static void print_value(param_t *param, const char *msg, ...)
 	va_end(va_args);
 }
 
+static void print_button_value(param_t *param, int n, const char *msg, ...) __attribute__((format(printf, 3, 4)));
+static void print_button_value(param_t *param, int n, const char *msg, ...)
+{
+	va_list va_args;
+	va_start(va_args, msg);
+
+	n++; /* Property is 0-indexed, X buttons are 1-indexed */
+	switch(param->printformat)
+	{
+		case FORMAT_XORG_CONF:
+			printf("Option \"%s%d\" \"", param->name, n);
+			vprintf(msg, va_args);
+			printf("\"\n");
+			break;
+		case FORMAT_SHELL:
+			printf("xsetwacom set \"%s\" \"%s\" \"%d\" \"",
+					param->device_name, param->name, n);
+			vprintf(msg, va_args);
+			printf("\"\n");
+			break;
+		case FORMAT_DEFAULT:
+		default:
+			vprintf(msg, va_args);
+			printf("\n");
+			break;
+	}
+
+	va_end(va_args);
+}
+
 static void usage(void)
 {
 	printf(
@@ -898,7 +954,7 @@ static void list(Display *dpy, int argc, char **argv)
 		 strcmp(argv[0], "modifiers") == 0)
 		list_mod(dpy);
 	else
-		printf("unknown argument to list.\n");
+		fprintf(stderr, "unknown argument to list.\n");
 }
 
 /**
@@ -908,7 +964,7 @@ static void list(Display *dpy, int argc, char **argv)
  * specialkeys.
  * @return The X Keysym representing specialkey.
  */
-static char *convert_specialkey(const char *specialkey)
+static const char *convert_specialkey(const char *specialkey)
 {
 	struct modifier *m = modifiers;
 
@@ -1071,12 +1127,41 @@ static int special_map_button(Display *dpy, int argc, char **argv, unsigned long
 	return i;
 }
 
-/* Return the first keycode to have the required keysym in the current group.
-   TODOs:
-   - parse other groups as well (do we need this?)
-   - for keysyms not on level 0, return the keycodes for the modifiers as
-     well
-*/
+/**
+ * Return the effective group number for a given keycode and requested
+ * global group number. If the group number is in-range for the keycode,
+ * the returned value will be identical. If the group number is out of
+ * range, the returned value will be wrapped, clamped, or redirected as
+ * necessary.
+ */
+static int keycode_effective_group(XkbDescPtr xkb, int keycode, int group)
+{
+	unsigned char info = XkbKeyGroupInfo(xkb, keycode);
+	unsigned char action = XkbOutOfRangeGroupInfo(info);
+	int n = XkbKeyNumGroups(xkb, keycode);
+
+	if (n <= 0)
+		return 0;
+	else if (group < n && group >= 0)
+		return group;
+	else if (action == XkbRedirectIntoRange)
+		return XkbOutOfRangeGroupNumber(info);
+	else if (action == XkbClampIntoRange)
+		return group < 1 ? 0 : group-1;
+	else /* if XkbWrapIntoRange */
+		return group % n;
+}
+
+/**
+ * Return the first keycode to have the required keysym. The effective
+ * group number is used durring scanning to ensure that an appropriate
+ * keycode is returned even if the active group does not define all
+ * keycodes.
+ *
+ *  TODOs:
+ *  - for keysyms not on level 0, return the keycodes for the modifiers as
+ *    well
+ */
 static int keysym_to_keycode(Display *dpy, KeySym sym)
 {
 	static XkbDescPtr xkb = NULL;
@@ -1090,10 +1175,11 @@ static int keysym_to_keycode(Display *dpy, KeySym sym)
 
 	for (kc = xkb->min_key_code; kc <= xkb->max_key_code; kc++)
 	{
+		int group = keycode_effective_group(xkb, kc, state.group);
 		int i;
 
-		for (i = 0; i < XkbKeyGroupWidth(xkb, kc, state.group); i++)
-			if (XkbKeycodeToKeysym(dpy, kc, state.group, i) == sym)
+		for (i = 0; i < XkbKeyGroupWidth(xkb, kc, group); i++)
+			if (XkbKeycodeToKeysym(dpy, kc, group, i) == sym)
 				goto out;
 	}
 
@@ -1115,7 +1201,7 @@ static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned
 		KeySym ks;
 		KeyCode kc;
 		int need_press = 0, need_release = 0;
-		char *key = argv[i];
+		const char *key = argv[i];
 
 		if (strlen(key) > 1)
 		{
@@ -1199,7 +1285,7 @@ static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned
 static char** strjoinsplit(int argc, char **argv, int *nwords)
 {
 	char **words = NULL;
-	int i, n;
+	int i;
 
 	*nwords = 0;
 	for (i = 0; i < argc; i++) {
@@ -1246,25 +1332,18 @@ static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* da
 	int  i = 0;
 	int  nwords = 0;
 	char **words = NULL;
-
+	
 	/* translate cmdline commands */
 	words = strjoinsplit(argc, argv, &nwords);
 
 	if (nwords==1 && sscanf(words[0], "%d", &i) == 1)
 	{ /* Mangle "simple" button maps into proper actions */
-		char **new_words = realloc(words, sizeof(char*)*2);
-		if (new_words == NULL)
-		{
-			fprintf(stderr, "Unable to reallocate memory.\n");
-			return False;
-		}
-
-		sprintf(new_words[0], "+%d", i);
-		new_words[1] = new_words[0];
-		new_words[0] = "button";
-
-		words  = new_words;
-		nwords = 2;
+		char *nargv[1];
+		
+		free(words);
+		nargv[0] = alloca(32);
+		sprintf(nargv[0], "button +%d", i);
+		words = strjoinsplit(1, nargv, &nwords);
 	}
 
 	for (i = 0; i < nwords && *nitems < size; i++)
@@ -1509,7 +1588,7 @@ static void set_mode(Display *dpy, XDevice *dev, param_t* param, int argc, char
 		mode = Absolute;
 	else
 	{
-		printf("Invalid device mode. Use 'Relative' or 'Absolute'.\n");
+		fprintf(stderr, "Invalid device mode. Use 'Relative' or 'Absolute'.\n");
 		return;
 	}
 
@@ -1651,7 +1730,7 @@ static void set(Display *dpy, int argc, char **argv)
 	dev = find_device(dpy, argv[0]);
 	if (!dev)
 	{
-		printf("Cannot find device '%s'.\n", argv[0]);
+		fprintf(stderr, "Cannot find device '%s'.\n", argv[0]);
 		return;
 	}
 
@@ -1660,11 +1739,11 @@ static void set(Display *dpy, int argc, char **argv)
 	{
 		if (is_deprecated_parameter(argv[1]))
 			goto out;
-		printf("Unknown parameter name '%s'.\n", argv[1]);
+		fprintf(stderr, "Unknown parameter name '%s'.\n", argv[1]);
 		goto out;
 	} else if (param->prop_flags & PROP_FLAG_READONLY)
 	{
-		printf("'%s' is a read-only option.\n", argv[1]);
+		fprintf(stderr, "'%s' is a read-only option.\n", argv[1]);
 		goto out;
 	}
 
@@ -1673,7 +1752,7 @@ static void set(Display *dpy, int argc, char **argv)
 		prop = XInternAtom(dpy, param->prop_name, True);
 		if (!prop || !test_property(dpy, dev, prop))
 		{
-			printf("Property '%s' does not exist on device.\n",
+			fprintf(stderr, "Property '%s' does not exist on device.\n",
 				param->prop_name);
 			goto out;
 		}
@@ -1790,7 +1869,7 @@ static void get_mode(Display *dpy, XDevice *dev, param_t* param, int argc, char
 
 static void get_rotate(Display *dpy, XDevice *dev, param_t* param, int argc, char **argv)
 {
-	char *rotation = NULL;
+	const char *rotation = NULL;
 	Atom prop, type;
 	int format;
 	unsigned char* data;
@@ -1942,7 +2021,10 @@ static int get_actions(Display *dpy, XDevice *dev,
 
 	XFree(data);
 
-	print_value(param, "%s", buff);
+	if (param->printformat == FORMAT_XORG_CONF)
+		fprintf(stderr, "%s: Actions are not supported by xorg.conf. Try shell format (-s) instead.\n", param->name);
+	else
+		print_button_value(param, offset, "%s", buff);
 
 	return 1;
 }
@@ -1988,7 +2070,7 @@ static int get_button(Display *dpy, XDevice *dev, param_t *param, int offset)
 		return 0;
 	}
 
-	print_value(param, "%d", prop);
+	print_button_value(param, offset, "%ld", prop);
 
 	return 1;
 }
@@ -2107,7 +2189,6 @@ static Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height,
 	unsigned long nitems, bytes_after;
 	float *data;
 	Bool matrix_is_valid = True;
-	int i;
 
 	int display_width = DisplayWidth(dpy, DefaultScreen(dpy));
 	int display_height = DisplayHeight(dpy, DefaultScreen(dpy));
@@ -2136,27 +2217,17 @@ static Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height,
 	TRACE("	[ %f %f %f ]\n", data[3], data[4], data[5]);
 	TRACE("	[ %f %f %f ]\n", data[6], data[7], data[8]);
 
-	for (i = 0; i < nitems && matrix_is_valid; i++)
-	{
-		switch (i) {
-			case 0: *width  = rint(display_width  * data[i]); break;
-			case 2: *x_org  = rint(display_width  * data[i]); break;
-			case 4: *height = rint(display_height * data[i]); break;
-			case 5: *y_org  = rint(display_height * data[i]); break;
-			case 8:
-				if (data[i] != 1)
-					matrix_is_valid = False;
-				break;
-			default:
-				if (data[i] != 0)
-					matrix_is_valid = False;
-				break;
-		}
-	}
-	XFree(data);
-
-	if (!matrix_is_valid)
+	*width  = rint(display_width  * data[0]);
+	*x_org  = rint(display_width  * data[2]);
+	*height = rint(display_height * data[4]);
+	*y_org  = rint(display_height * data[5]);
+	if ((data[1] != 0 || data[3] != 0 || data[6] != 0 || data[7] != 0) ||
+	    (data[8] != 1)) {
 		fprintf(stderr, "Non-rectangular transformation matrix detected.\n");
+		matrix_is_valid = False;
+	}
+
+	XFree(data);
 
 	return matrix_is_valid;
 }
@@ -2307,7 +2378,7 @@ static Bool set_output_xrandr(Display *dpy, XDevice *dev, char *output_name)
 	XRRFreeScreenResources(res);
 
 	if (!success)
-		printf("Unable to find output '%s'. "
+		fprintf(stderr, "Unable to find output '%s'. "
 			"Output may not be connected.\n", output_name);
 
 	return success;
@@ -2483,7 +2554,27 @@ static void get_all(Display *dpy, XDevice *dev, param_t *param, int argc, char *
 		{
 			p->device_name = param->device_name;
 			p->printformat = param->printformat;
-			get_param(dpy, dev, p, argc, argv);
+
+			if (p->prop_name && strcmp(p->prop_name, WACOM_PROP_BUTTON_ACTIONS) == 0)
+			{
+				int i;
+				int nmap = 256;
+				unsigned char map[nmap];
+
+				nmap = XGetDeviceButtonMapping(dpy, dev, map, nmap);
+				for (i = 1; i <= nmap; i++)
+				{
+					char tmparg[16];
+					char *tmpargv = &tmparg[0];
+
+					if (i > 3 && i < 8)
+						continue;
+					sprintf(tmparg, "%d", i);
+					get_param(dpy, dev, p, 1, &tmpargv);
+				}
+			}
+			else
+				get_param(dpy, dev, p, argc, argv);
 		}
 		p++;
 	}
@@ -2505,7 +2596,7 @@ static void get(Display *dpy, enum printformat printformat, int argc, char **arg
 	dev = find_device(dpy, argv[0]);
 	if (!dev)
 	{
-		printf("Cannot find device '%s'.\n", argv[0]);
+		fprintf(stderr, "Cannot find device '%s'.\n", argv[0]);
 		return;
 	}
 
@@ -2514,11 +2605,11 @@ static void get(Display *dpy, enum printformat printformat, int argc, char **arg
 	{
 		if (is_deprecated_parameter(argv[1]))
 			return;
-		printf("Unknown parameter name '%s'.\n", argv[1]);
+		fprintf(stderr, "Unknown parameter name '%s'.\n", argv[1]);
 		return;
 	} else if (param->prop_flags & PROP_FLAG_WRITEONLY)
 	{
-		printf("'%s' is a write-only option.\n", argv[1]);
+		fprintf(stderr, "'%s' is a write-only option.\n", argv[1]);
 		return;
 	} else
 	{
@@ -2545,7 +2636,7 @@ static void get_param(Display *dpy, XDevice *dev, param_t *param, int argc, char
 		prop = XInternAtom(dpy, param->prop_name, True);
 		if (!prop || !test_property(dpy, dev, prop))
 		{
-			printf("Property '%s' does not exist on device.\n",
+			fprintf(stderr, "Property '%s' does not exist on device.\n",
 				param->prop_name);
 			return;
 		}
@@ -2683,7 +2774,7 @@ int main (int argc, char **argv)
 	dpy = XOpenDisplay(display);
 	if (!dpy)
 	{
-		printf("Failed to open Display %s.\n", display ? display : "");
+		fprintf(stderr, "Failed to open Display %s.\n", display ? display : "");
 		return -1;
 	}
 
@@ -2761,7 +2852,7 @@ static void test_is_modifier(void)
 static void test_convert_specialkey(void)
 {
 	char i;
-	char *converted;
+	const char *converted;
 	char buff[5];
 	struct modifier *m;
 
